from time import time
from functools import wraps
import logging

logger = logging.getLogger(__name__)


def measure_timing(func):
    @wraps(func)
    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        logger.info(f'flow execution completes in {(t2-t1):.4f}s')
        return result
    return wrap_func



from service_flow.middleware import Middleware, DecoratorMiddleware
from service_flow.flow import Flow
from service_flow.exceptions import ForkException
import pytest
import inspect

class InplaceModification2(Middleware):
    def __init__(self, increment):
        self.increment = increment

    async def __call__(self, bar: list):
        return await {'bar': [i + self.increment for i in bar]}


class AddContextVariable2(Middleware):
    async def __call__(self, foo, bar):
        return {'baz': 13}


class ReturnNone2(Middleware):
    async def __call__(self, baz):
        ...

class NestedService2(DecoratorMiddleware):
    async def __call__(self, context):
        try:
            return await self.app(context)
        except ZeroDivisionError:
            return {'error': 'decided by zero'}

@pytest.mark.asyncio
async def test_nested():
    flow = NestedService2() >> AddContextVariable2()
    assert type(flow) is Flow
    assert flow.async_mode == True
    assert await flow({'foo': 1, 'bar': 2}) == {'bar': 2, 'baz': 13, 'foo': 1}


@pytest.mark.asyncio
async def test_start_with_async():
    flow = AddContextVariable2() >> ReturnNone2()
    assert type(flow) is Flow
    assert flow.async_mode == True
    assert await flow({'foo': 1, 'bar': 2}) == {'foo': 1, 'bar': 2, 'baz': 13}


@pytest.mark.asyncio
async def test_multi_levels():
    flow = InplaceModification2(1) >>  \
            AddContextVariable2() >>  \
            ReturnNone2()

    assert type(flow) is Flow
    assert [type(m[0]) for m in flow.middlewares] == [
        InplaceModification2, AddContextVariable2, ReturnNone2
    ]

    assert [m[1] for m in flow.middlewares] == [
        ['bar'], ['foo', 'bar'], ['baz']
    ]

    assert flow.async_mode == True

    result = await flow({'foo': 'a', 'bar': [1, 2]})

    assert result == {
        'foo': 'a',
        'bar': [2, 3],
        'baz': 13
    }


class SetScenario(Middleware):
    def __call__(self, scenario: str):
        return {'scenario': scenario}

class OutputScenario(Middleware):
    def __call__(self, scenario):
        return {'scenario': scenario}

class InplaceModification2(Middleware):
    def __init__(self, increment):
        self.increment = increment

    def __call__(self, bar: list):
        return {'bar': [i + self.increment for i in bar]}


def test_fork3():
    flow = SetScenario() < \
            ('scenario', {
                'scenario1': InplaceModification2(1) >> OutputScenario(),
                'scenario2': InplaceModification2(2) >> OutputScenario(),
                })
    assert flow({'scenario': 'scenario1', 'foo': 1, 'bar': [1, 2]}) == {'scenario': 'scenario1', 'foo': 1, 'bar': [2, 3]}
    assert flow({'scenario': 'scenario2', 'foo': 2, 'bar': [1, 2]}) == {'scenario': 'scenario2', 'foo': 2, 'bar': [3, 4]}

    with pytest.raises(ForkException):
        assert flow({'scenario': 'scenario3', 'foo': 1, 'bar': [1, 2]})



import itertools
from service_flow.exceptions import ForkException


class Fork():
    def __init__(self, fork_var: str, fork_conditions: dict):
        self.fork_var = fork_var
        self.fork_conditions = fork_conditions

    def __call__(self, **context):
        value = context[self.fork_var]
        try:
            func = self.fork_conditions[value]
        except KeyError as e:
            raise ForkException(f'value {value} does not match any fork key') from e
        return func(context)

    @property
    def uniq_arguments(self):
        list2d = [middleware[1] for _, flow in self.fork_conditions.items() for middleware in flow.middlewares]
        list2d.append([self.fork_var])
        return list(set(itertools.chain(*list2d)))



from service_flow.middleware import Middleware
from service_flow.exceptions import StopFlowException
import pytest

class StopFlow(Middleware):
    def __call__(self):
        raise StopFlowException('some error message')


def test_stopflow():
    flow = StopFlow() >>  \
            (lambda foo: {'return_value': foo})
    with pytest.raises(StopFlowException):
        flow({})



from service_flow.middleware import Middleware, DecoratorMiddleware
from service_flow.flow import Flow
from service_flow.exceptions import ForkException
import pytest


class InplaceModification(Middleware):
    def __init__(self, increment):
        self.increment = increment

    def __call__(self, bar: list):
        return {'bar': [i + self.increment for i in bar]}


class AddContextVariable(Middleware):
    def __call__(self, foo, bar):
        return {'baz': 13}


class ReturnNone(Middleware):
    def __call__(self, baz):
        ...


class SetScenario(Middleware):
    def __call__(self, scenario: str):
        return {'scenario': scenario}


def test_multi_levels():
    flow = InplaceModification(1) >>  \
            AddContextVariable() >>  \
            ReturnNone()

    assert type(flow) is Flow
    assert [type(m[0]) for m in flow.middlewares] == [
        InplaceModification, AddContextVariable, ReturnNone
    ]

    assert [m[1] for m in flow.middlewares] == [
        ['bar'], ['foo', 'bar'], ['baz']
    ]

    assert flow({'foo': 'a', 'bar': [1, 2]}) == {
        'foo': 'a',
        'bar': [2, 3],
        'baz': 13
    }


def test_lambda():
    flow = InplaceModification(1) >>  \
            (lambda foo: {'return_value': foo})

    assert flow({'foo': 'a', 'bar': [1, 2]}) == {
        'foo': 'a',
        'bar': [2, 3],
        'return_value': 'a'
    }

def test_fork3():
    flow = SetScenario() < \
            ('scenario', {
                'scenario1': InplaceModification(1) >> AddContextVariable(),
                'scenario2': InplaceModification(2) >> AddContextVariable(),
                })
    assert flow({'scenario': 'scenario1', 'foo': 1, 'bar': [1, 2]}) == {'scenario': 'scenario1', 'foo': 1, 'bar': [2, 3], 'baz': 13}
    assert flow({'scenario': 'scenario2', 'foo': 1, 'bar': [1, 2]}) == {'scenario': 'scenario2', 'foo': 1, 'bar': [3, 4], 'baz': 13}

    with pytest.raises(ForkException):
        assert flow({'scenario': 'scenario3', 'foo': 1, 'bar': [1, 2]})

# def test_fork():
#     flow = SetScenario() < \
#             ('scenario', {
#                 'scenario1': lambda : {'results': '1'},
#                 'scenario2': lambda : {'results': '2'},
#                 })
#     assert flow({'scenario': 'scenario1'}) == {'results': '1', 'scenario': 'scenario1'}
#     assert flow({'scenario': 'scenario2'}) == {'results': '2', 'scenario': 'scenario2'}


# def test_fork2():
#     flow = SetScenario() >> \
#             (lambda scenario : {'value': scenario}) < \
#             ('scenario', {
#                 'scenario1': lambda : {'results': '1'},
#                 'scenario2': lambda : {'results': '2'},
#                 })
#     assert flow({'scenario': 'scenario1'}) == {'results': '1', 'scenario': 'scenario1', 'value': 'scenario1'}
#     assert flow({'scenario': 'scenario2'}) == {'results': '2', 'scenario': 'scenario2', 'value': 'scenario2'}


class NestedService(DecoratorMiddleware):
    def __call__(self, context):
        try:
            self.app(context)
        except ZeroDivisionError:
            return {'error': 'decided by zero'}

class CustomException(Exception):
    pass

def test_nested():
    flow = InplaceModification(1) >>  \
            NestedService() >>  \
            (lambda foo: {'result': 1/foo})
    assert flow({'foo': 0, 'bar': []}) == {
        'bar': [],
        'error': 'decided by zero',
        'foo': 0
    }

    assert flow({'foo': 1, 'bar': []}) == {
        'bar': [],
        'foo': 1,
        'result': 1
    }

    flow = NestedService() >>  \
            (lambda foo: {'result': 1/foo})
    assert flow({'foo': 0, 'bar': []}) == {
        'bar': [],
        'error': 'decided by zero',
        'foo': 0
    }

    flow = NestedService() >> (lambda foo: {'result1': 1/foo}) >> (lambda bar: {'result2': 2/bar})
    assert flow({'foo': 1, 'bar': 2}) == {
        'bar': 2,
        'foo': 1,
        'result1': 1,
        'result2': 1
    }

    class DoNothing(Middleware):
        def __call__(self):
            return {}

    flow = DoNothing() >> NestedService() >> (lambda foo: {'result1': 1/foo}) >> (lambda bar: {'result2': 2/bar})
    assert flow({'foo': 1, 'bar': 2}) == {
        'bar': 2,
        'foo': 1,
        'result1': 1,
        'result2': 1
    }


import inspect
from types import LambdaType
from typing import Callable
from service_flow.fork import Fork
from service_flow.timer import measure_timing
import logging
logger = logging.getLogger(__name__)


class Flow():
    def __init__(self, m1: Callable):
        self.middlewares = []
        self.last_flow = self
        self.async_mode = False
        self._add_middleware(m1)

    @measure_timing
    def __call__(self, context: dict={}):
        return self.__run__(context) if self.async_mode == False else self.__async_run__(context)

    def __run__(self, context: dict):
        for middleware, kw_nms in self.middlewares:
            kwargs = self.middleware_kwargs(context, kw_nms)
            context_mods = middleware(**kwargs)
            context.update(context_mods if isinstance(context_mods, dict) else {})
        return context

    @staticmethod
    def middleware_kwargs(context: dict, kw_nms: list):
        if kw_nms == ['context']:  # context is a reserved argument keyword
            # if the middleware has a "context" parameter, then pass the entire context
            return {'context': context}
        else:
            # calls the middleware and get the result as modifications to the context
            return {key: context[key] for key in kw_nms if key in context}

    async def __async_run__(self, context: dict):
        for middleware, kw_nms in self.middlewares:
            kwargs = self.middleware_kwargs(context, kw_nms)
            # get the middleware output a la sync/async
            context_mods = await middleware(**kwargs) if middleware.is_async else middleware(**kwargs)
            context.update(context_mods if isinstance(context_mods, dict) else {})
        return context

    # handle >> operator for the sequential flow
    def __rshift__(self, middleware):
        # if the last middleware is a DecoratorMiddleware,
        # then it is a nested flow, and we need to add the new middleware to the nested flow
        if hasattr(self.last_middleware, 'app'):
            # for decorator pattern, the next middleware is added to the nested flow
            # so we are creating the new flow and keep its as an instance variable, and
            # next middleware will be added to this flow instead of the original one
            flow = Flow(middleware)
            self.last_middleware.app = flow
            self.last_flow = flow
        else:
            self.last_flow._add_middleware(middleware)

        # always return self, so that we can chain the >> operator
        return self

    @property
    def last_middleware(self):
        return self.last_flow.middlewares[-1][0]

    # handle < operator for the fork flow
    def __lt__(self, conditions: tuple):
        self.add_fork(conditions)
        return self

    def _add_middleware(self, middleware):
        # if middleware is an async function, then set the async_mode to True
        if not self.async_mode and \
            (not isinstance(middleware, LambdaType)) and middleware.is_async:
            self.async_mode = True

        self.middlewares.append((middleware, Flow.arguments(middleware)))

    def add_fork(self, conditions: tuple):
        variable_nm = conditions[0]
        variable_conditions = conditions[1]
        fork = Fork(variable_nm, variable_conditions)
        self.middlewares.append((fork, fork.uniq_arguments))
        return self

    @staticmethod
    def arguments(middleware):
        if isinstance(middleware, LambdaType):  # support for lambda func
            return inspect.getfullargspec(middleware).args
        else:  # Middleware object
            return inspect.getfullargspec(middleware.__call__).args[1:]



class StopFlowException(Exception):
    """
    an exception to stop the execution of subsequent middlewares
    """


class RetryException(Exception):
    """
    an exception to stop the execution of subsequent middlewares and signal for re-process the same input overall again
    """


class FatalException(Exception):
    """
    an exception to stop the execution of subsequent middlewares and for the processor to exit and start fresh
    """


class ForkException(FatalException):
    """
    an exception to stop the execution of subsequent middlewares and log the exception as an error
    """



from service_flow.flow import Flow
import inspect
from functools import cached_property


class Middleware():
    # the following methods are only called on the first middleware,
    # either at the beginning of the flow or nested flow
    def __rshift__(self, middleware):
        s = Flow(self)
        s.__rshift__(middleware)
        return s

    def __lt__(self, conditions: tuple):
        s = Flow(self)
        s.add_fork(conditions)
        return s

    @cached_property
    def is_async(self):
        return inspect.iscoroutinefunction(self.__call__)

class DecoratorMiddleware(Middleware):
    def __init__(self):
        self.app: Flow = None
